package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.BufferUploader
import com.mojang.blaze3d.vertex.DefaultVertexFormat
import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.blaze3d.vertex.VertexFormat
import net.minecraft.client.gui.Font
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.network.chat.Component
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kommons.math.linearInterpolation
import kotlin.math.PI
import kotlin.math.acos
import kotlin.math.cos
import kotlin.math.pow
import kotlin.math.sign
import kotlin.math.sin

private const val LINE_WIDTH = 1f
private const val HIGHLIGHT_WIDTH = 3f

private val TEXT_BACKGROUND = RGBAColor(0f, 0f, 0f, 0.5f)

data class GraphMouseLabels(
	val mouseX: Float,
	val mouseY: Float,
	val labels: (Int) -> Component,
	val font: Font,
	val color: RGBAColor = RGBAColor.WHITE,
	val pillarColor: RGBAColor = RGBAColor.WHITE,
	val textColor: RGBAColor = RGBAColor.WHITE,
	val textBackgroundColor: RGBAColor = TEXT_BACKGROUND
)

fun renderGraph(
	poseStack: PoseStack,
	normalized: FloatArray,
	width: Float,
	height: Float,
	color: RGBAColor = RGBAColor.WHITE,
	labels: GraphMouseLabels? = null,
	x: Float = 0f,
	y: Float = 0f,
) {
	val builder = tesselator.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION)
	val step = width / normalized.size

	val pose = poseStack.last().pose
	val checkLabels = labels != null && labels.mouseY in 0f .. height && labels.mouseX in 0f .. width

	var drawLabel = -1
	var drawPointX = 0f
	var drawPointY = 0f

	for (i in 0 until normalized.size - 1) {
		val y0 = y + (1f - normalized[i]) * height
		val y1 = y + (1f - normalized[i + 1]) * height

		val x0 = x + width - (i + 0.5f) * step
		val x1 = x + width - (i + 1.5f) * step

		val xDiff = x1 - x0
		val yDiff = y1 - y0

		val length = (xDiff.pow(2f) + yDiff.pow(2f)).pow(0.5f)
		val angle = acos(xDiff / length) * yDiff.sign - PI / 2f // rotate 90 deg

		val xDisp = cos(angle).toFloat() * LINE_WIDTH / 2f
		val yDisp = sin(angle).toFloat() * LINE_WIDTH / 2f

		builder.vertex(pose, x1 - xDisp, y1 + yDisp, 0f)
		builder.vertex(pose, x1 + xDisp, y1 - yDisp, 0f)
		builder.vertex(pose, x0 + xDisp, y0 - yDisp, 0f)
		builder.vertex(pose, x0 - xDisp, y0 + yDisp, 0f)

		//graphics.renderRect(x0, y0, LINE_WIDTH, LINE_WIDTH)

		if (checkLabels && drawLabel == -1 && labels!!.mouseX in x1 .. x0) {
			drawLabel = i
			drawPointX = x0 // linearInterpolation(0.5f, x0, x1)
			drawPointY = y0 // linearInterpolation(0.5f, y0, y1)
		}
	}

	RenderSystem.setShader(GameRenderer::getPositionShader)
	RenderSystem.setShaderColor(color.red, color.green, color.blue, color.alpha)
	RenderSystem.disableDepthTest()
	RenderSystem.disableBlend()

	BufferUploader.drawWithShader(builder.buildOrThrow())

	if (drawLabel != -1) {
		renderRect(pose, drawPointX - LINE_WIDTH / 2f, y, LINE_WIDTH, height, color = labels!!.pillarColor)
		renderRect(pose, drawPointX - HIGHLIGHT_WIDTH / 2f, drawPointY - HIGHLIGHT_WIDTH / 2f, HIGHLIGHT_WIDTH, HIGHLIGHT_WIDTH, color = labels.color)

		val label = labels.labels(drawLabel)
		val fWidth = labels.font.width(label).toFloat() + 2f
		val fHeight = labels.font.lineHeight.toFloat() + 2f

		val anchorX = labels.mouseX // drawPointX
		val anchorY = labels.mouseY - fHeight / 2f - 2f

		renderRect(pose, anchorX - fWidth / 2f, anchorY - fHeight / 2f, fWidth, fHeight, color = labels.textBackgroundColor)
		labels.font.draw(poseStack, label, anchorX, anchorY, gravity = RenderGravity.CENTER_CENTER, color = labels.textColor)
	}
}
